@@ -3,44 +3,10 @@
 #include "frameobject.h"
 
 #include <ffi.h>
-#include "ctypes.h"
-
 #ifdef MS_WIN32
 #include <windows.h>
 #endif
-
-/* For 2.3, use the PyGILState_ calls, see PEP 311 */
-#if (PY_VERSION_HEX >= 0x02030000)
-#define CTYPES_USE_GILSTATE
-#endif
-
-#ifndef CTYPES_USE_GILSTATE
-static PyInterpreterState *g_interp;	/* need this to create new thread states */
-
-static void EnterPython(void)
-{
-	PyThreadState *pts;
-	if (!g_interp)
-		Py_FatalError("_ctypes: no interpreter state");
-	PyEval_AcquireLock();
-	pts = PyThreadState_New(g_interp);
-	if (!pts)
-		Py_FatalError("_ctypes: Could not allocate ThreadState");
-	if (NULL != PyThreadState_Swap(pts))
-		Py_FatalError("_ctypes (EnterPython): thread state not == NULL?");
-}
-
-static void LeavePython(void)
-{
-	PyThreadState *pts = PyThreadState_Get();
-	if (!pts)
-		Py_FatalError("_ctypes (LeavePython): ThreadState is NULL?");
-	PyThreadState_Clear(pts);
-	pts = PyThreadState_Swap(NULL);
-	PyThreadState_Delete(pts);
-	PyEval_ReleaseLock();
-}
-#endif
+#include "ctypes.h"
 
 static void
 PrintError(char *msg, ...)
@@ -157,12 +123,7 @@ static void _CallPythonObject(void *mem,
 	PyObject *result;
 	PyObject *arglist = NULL;
 	int nArgs;
-
-#ifdef CTYPES_USE_GILSTATE
 	PyGILState_STATE state = PyGILState_Ensure();
-#else
-	EnterPython();
-#endif
 
 	nArgs = PySequence_Length(converters);
 	/* Hm. What to return in case of error?
@@ -189,10 +150,11 @@ static void _CallPythonObject(void *mem,
 			goto Done;
 		}
 
-		if (dict && dict->getfunc) {
+		if (dict && dict->getfunc && !IsSimpleSubType(cnv)) {
 			PyObject *v = dict->getfunc(*pArgs, dict->size);
 			if (!v) {
 				PrintError("create argument %d:\n", i);
+				Py_DECREF(cnv);
 				goto Done;
 			}
 			PyTuple_SET_ITEM(arglist, i, v);
@@ -206,10 +168,12 @@ static void _CallPythonObject(void *mem,
 			CDataObject *obj = (CDataObject *)PyObject_CallFunctionObjArgs(cnv, NULL);
 			if (!obj) {
 				PrintError("create argument %d:\n", i);
+				Py_DECREF(cnv);
 				goto Done;
 			}
 			if (!CDataObject_Check(obj)) {
 				Py_DECREF(obj);
+				Py_DECREF(cnv);
 				PrintError("unexpected result of create argument %d:\n", i);
 				goto Done;
 			}
@@ -222,8 +186,10 @@ static void _CallPythonObject(void *mem,
 			PyErr_SetString(PyExc_TypeError,
 					"cannot build parameter");
 			PrintError("Parsing argument %d\n", i);
+			Py_DECREF(cnv);
 			goto Done;
 		}
+		Py_DECREF(cnv);
 		/* XXX error handling! */
 		pArgs++;
 	}
@@ -233,7 +199,8 @@ if (x == NULL) _AddTraceback(what, __FILE__, __LINE__ - 1), PyErr_Print()
 
 	result = PyObject_CallObject(callable, arglist);
 	CHECK("'calling callback function'", result);
-	if (result && result != Py_None) { /* XXX What is returned for Py_None ? */
+	if ((restype != &ffi_type_void)
+	    && result && result != Py_None) { /* XXX What is returned for Py_None ? */
 		/* another big endian hack */
 		union {
 			char c;
@@ -242,6 +209,7 @@ if (x == NULL) _AddTraceback(what, __FILE__, __LINE__ - 1), PyErr_Print()
 			long l;
 		} r;
 		PyObject *keep;
+		assert(setfunc);
 		switch (restype->size) {
 		case 1:
 			keep = setfunc(&r, result, 0);
@@ -270,32 +238,32 @@ if (x == NULL) _AddTraceback(what, __FILE__, __LINE__ - 1), PyErr_Print()
 			CHECK("'converting callback result'", keep);
 			break;
 		}
-		/* assert (keep == Py_None); */
-		/* XXX We have no way to keep the needed reference XXX */
-		/* Should we emit a warning? */
-		Py_XDECREF(keep);
+		/* keep is an object we have to keep alive so that the result
+		   stays valid.  If there is no such object, the setfunc will
+		   have returned Py_None.
+
+		   If there is such an object, we have no choice than to keep
+		   it alive forever - but a refcount and/or memory leak will
+		   be the result.  EXCEPT when restype is py_object - Python
+		   itself knows how to manage the refcount of these objects.
+		*/
+		if (keep == NULL) /* Could not convert callback result. */
+			PyErr_WriteUnraisable(Py_None);
+		else if (keep == Py_None) /* Nothing to keep */
+			Py_DECREF(keep);
+		else if (setfunc != getentry("O")->setfunc) {
+			if (-1 == PyErr_Warn(PyExc_RuntimeWarning,
+					     "memory leak in callback function."))
+				PyErr_WriteUnraisable(Py_None);
+		}
 	}
 	Py_XDECREF(result);
   Done:
 	Py_XDECREF(arglist);
 	
-#ifdef CTYPES_USE_GILSTATE
 	PyGILState_Release(state);
-#else
-	LeavePython();
-#endif
 }
 
-typedef struct {
-	ffi_closure *pcl; /* the C callable */
-	ffi_cif cif;
-	PyObject *converters;
-	PyObject *callable;
-	SETFUNC setfunc;
-	ffi_type *restype;
-	ffi_type *atypes[0];
-} ffi_info;
-
 static void closure_fcn(ffi_cif *cif,
 			void *resp,
 			void **args,
@@ -311,17 +279,10 @@ static void closure_fcn(ffi_cif *cif,
 			  args);
 }
 
-void FreeCallback(THUNK thunk)
-{
-	FreeClosure(*(void **)thunk);
-	PyMem_Free(thunk);
-}
-
-THUNK AllocFunctionCallback(PyObject *callable,
-			    int nArgBytes,
-			    PyObject *converters,
-			    PyObject *restype,
-			    int is_cdecl)
+ffi_info *AllocFunctionCallback(PyObject *callable,
+				PyObject *converters,
+				PyObject *restype,
+				int is_cdecl)
 {
 	int result;
 	ffi_info *p;
@@ -329,38 +290,39 @@ THUNK AllocFunctionCallback(PyObject *callable,
 	ffi_abi cc;
 
 	nArgs = PySequence_Size(converters);
-	p = (ffi_info *)PyMem_Malloc(sizeof(ffi_info) + sizeof(ffi_type) * nArgs);
+	p = (ffi_info *)PyMem_Malloc(sizeof(ffi_info) + sizeof(ffi_type) * (nArgs + 1));
 	if (p == NULL) {
 		PyErr_NoMemory();
 		return NULL;
 	}
 	p->pcl = MallocClosure();
 	if (p->pcl == NULL) {
-		PyMem_Free(p);
 		PyErr_NoMemory();
-		return NULL;
+		goto error;
 	}
 
 	for (i = 0; i < nArgs; ++i) {
 		PyObject *cnv = PySequence_GetItem(converters, i);
+		if (cnv == NULL)
+			goto error;
 		p->atypes[i] = GetType(cnv);
 		Py_DECREF(cnv);
 	}
 	p->atypes[i] = NULL;
 
-	{
+	if (restype == Py_None) {
+		p->setfunc = NULL;
+		p->restype = &ffi_type_void;
+	} else {
 		StgDictObject *dict = PyType_stgdict(restype);
-		if (dict) {
-			p->setfunc = dict->setfunc;
-			p->restype = &dict->ffi_type;
-		} else {
-			p->setfunc = getentry("i")->setfunc;
-			p->restype = &ffi_type_sint;
-		}
+		if (dict == NULL)
+			goto error;
+		p->setfunc = dict->setfunc;
+		p->restype = &dict->ffi_type_pointer;
 	}
 
 	cc = FFI_DEFAULT_ABI;
-#ifdef MS_WIN32
+#if defined(MS_WIN32) && !defined(_WIN32_WCE)
 	if (is_cdecl == 0)
 		cc = FFI_STDCALL;
 #endif
@@ -370,21 +332,26 @@ THUNK AllocFunctionCallback(PyObject *callable,
 	if (result != FFI_OK) {
 		PyErr_Format(PyExc_RuntimeError,
 			     "ffi_prep_cif failed with %d", result);
-		PyMem_Free(p);
-		return NULL;
+		goto error;
 	}
 	result = ffi_prep_closure(p->pcl, &p->cif, closure_fcn, p);
 	if (result != FFI_OK) {
 		PyErr_Format(PyExc_RuntimeError,
 			     "ffi_prep_closure failed with %d", result);
-		PyMem_Free(p);
-		return NULL;
+		goto error;
 	}
 
 	p->converters = converters;
 	p->callable = callable;
+	return p;
 
-	return (THUNK)p;
+  error:
+	if (p) {
+		if (p->pcl)
+			FreeClosure(p->pcl);
+		PyMem_Free(p);
+	}
+	return NULL;
 }
 
 /****************************************************************************
@@ -396,125 +363,15 @@ void init_callbacks_in_module(PyObject *m)
 {
 	if (PyType_Ready((PyTypeObject *)&PyType_Type) < 0)
 		return;
-
-#ifndef CTYPES_USE_GILSTATE
-	g_interp = PyThreadState_Get()->interp;
-#endif
 }
 
 #ifdef MS_WIN32
-/*
-   Modeled after a function from Mark Hammond.
-
-   Obtains a string from a Python traceback.  This is the exact same string as
-   "traceback.print_exception" would return.
-
-   Result is a string which must be free'd using PyMem_Free()
-*/
-#define TRACEBACK_FETCH_ERROR(what) {errMsg = what; goto done;}
-
-char *PyTraceback_AsString(void)
-{
-	char *errMsg = NULL; /* holds a local error message */
-	char *result = NULL; /* a valid, allocated result. */
-	PyObject *modStringIO = NULL;
-	PyObject *modTB = NULL;
-	PyObject *obStringIO = NULL;
-	PyObject *obResult = NULL;
-
-	PyObject *type, *value, *traceback;
-
-	PyErr_Fetch(&type, &value, &traceback);
-	PyErr_NormalizeException(&type, &value, &traceback);
-	
-	modStringIO = PyImport_ImportModule("cStringIO");
-	if (modStringIO==NULL)
-		TRACEBACK_FETCH_ERROR("cant import cStringIO\n");
-
-	obStringIO = PyObject_CallMethod(modStringIO, "StringIO", NULL);
-
-	/* Construct a cStringIO object */
-	if (obStringIO==NULL)
-		TRACEBACK_FETCH_ERROR("cStringIO.StringIO() failed\n");
-
-	modTB = PyImport_ImportModule("traceback");
-	if (modTB==NULL)
-		TRACEBACK_FETCH_ERROR("cant import traceback\n");
-
-	obResult = PyObject_CallMethod(modTB, "print_exception",
-				       "OOOOO",
-				       type, value ? value : Py_None,
-				       traceback ? traceback : Py_None,
-				       Py_None,
-				       obStringIO);
-				    
-	if (obResult==NULL) 
-		TRACEBACK_FETCH_ERROR("traceback.print_exception() failed\n");
-	Py_DECREF(obResult);
-
-	obResult = PyObject_CallMethod(obStringIO, "getvalue", NULL);
-	if (obResult==NULL) 
-		TRACEBACK_FETCH_ERROR("getvalue() failed.\n");
-
-	/* And it should be a string all ready to go - duplicate it. */
-	if (!PyString_Check(obResult))
-			TRACEBACK_FETCH_ERROR("getvalue() did not return a string\n");
-
-	{ // a temp scope so I can use temp locals.
-		char *tempResult = PyString_AsString(obResult);
-		result = (char *)PyMem_Malloc(strlen(tempResult)+1);
-		if (result==NULL)
-			TRACEBACK_FETCH_ERROR("memory error duplicating the traceback string\n");
-		strcpy(result, tempResult);
-	} // end of temp scope.
-done:
-	/* All finished - first see if we encountered an error */
-	if (result==NULL && errMsg != NULL) {
-		result = (char *)PyMem_Malloc(strlen(errMsg)+1);
-		if (result != NULL)
-			/* if it does, not much we can do! */
-			strcpy(result, errMsg);
-	}
-	Py_XDECREF(modStringIO);
-	Py_XDECREF(modTB);
-	Py_XDECREF(obStringIO);
-	Py_XDECREF(obResult);
-	Py_XDECREF(value);
-	Py_XDECREF(traceback);
-	Py_XDECREF(type);
-	return result;
-}
-
-void MyPyErr_Print(char *msg)
-{
-	char *text;
-
-	text = PyTraceback_AsString();
-	MessageBox(NULL,
-		   text,
-		   msg,
-		   MB_OK | MB_ICONSTOP);
-	PyMem_Free(text);
-}
 
 static void LoadPython(void)
 {
 	if (!Py_IsInitialized()) {
 		PyEval_InitThreads();
 		Py_Initialize();
-	} else {
-#ifndef CTYPES_USE_GILSTATE
-		/* Python is already initialized.
-		   We assume we don't have the lock.
-		*/
-		if (!g_interp) {
-			g_interp = PyInterpreterState_Head();
-			if(PyInterpreterState_Next(g_interp))
-				Py_FatalError("_ctypes: more than one interpreter state");
-		}
-		EnterPython();
-#endif
-		return;
 	}
 }
 
@@ -524,16 +381,22 @@ long Call_GetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppv)
 {
 	PyObject *mod, *func, *result;
 	long retval;
+	static PyObject *context;
 
-	mod = PyImport_ImportModule("ctypes.com.server");
-	if (!mod)
+	if (context == NULL)
+		context = PyString_FromString("_ctypes.DllGetClassObject");
+
+	mod = PyImport_ImportModule("ctypes");
+	if (!mod) {
+		PyErr_WriteUnraisable(context ? context : Py_None);
 		/* There has been a warning before about this already */
 		return E_FAIL;
+	}
 
 	func = PyObject_GetAttrString(mod, "DllGetClassObject");
 	Py_DECREF(mod);
 	if (!func) {
-		MyPyErr_Print("DllGetClassObject");
+		PyErr_WriteUnraisable(context ? context : Py_None);
 		return E_FAIL;
 	}
 
@@ -541,13 +404,15 @@ long Call_GetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppv)
 				       "iii", rclsid, riid, ppv);
 	Py_DECREF(func);
 	if (!result) {
-		MyPyErr_Print("Called DllGetClassObject");
+		PyErr_WriteUnraisable(context ? context : Py_None);
 		return E_FAIL;
 	}
 
 	retval = PyInt_AsLong(result);
-	if (PyErr_Occurred())
-		MyPyErr_Print("Convert result of DllGetClassObject to int");
+	if (PyErr_Occurred()) {
+		PyErr_WriteUnraisable(context ? context : Py_None);
+		retval = E_FAIL;
+	}
 	Py_DECREF(result);
 	return retval;
 }
@@ -557,21 +422,12 @@ STDAPI DllGetClassObject(REFCLSID rclsid,
 			 LPVOID *ppv)
 {
 	long result;
-
-#ifdef CTYPES_USE_GILSTATE
 	PyGILState_STATE state;
+
 	LoadPython();
 	state = PyGILState_Ensure();
-#else
-	LoadPython(); /* calls EnterPython itself */
-#endif
 	result = Call_GetClassObject(rclsid, riid, ppv);
-
-#ifdef CTYPES_USE_GILSTATE
 	PyGILState_Release(state);
-#else
-	LeavePython();
-#endif
 	return result;
 }
 
@@ -579,30 +435,39 @@ long Call_CanUnloadNow(void)
 {
 	PyObject *mod, *func, *result;
 	long retval;
+	static PyObject *context;
 
-	mod = PyImport_ImportModule("ctypes.com.server");
+	if (context == NULL)
+		context = PyString_FromString("_ctypes.DllCanUnloadNow");
+
+	mod = PyImport_ImportModule("ctypes");
 	if (!mod) {
-		MyPyErr_Print("Importing ctypes.com.server in Call_CanUnloadNow");
+/*		OutputDebugString("Could not import ctypes"); */
+		/* We assume that this error can only occur when shutting
+		   down, so we silently ignore it */
+		PyErr_Clear();
 		return E_FAIL;
 	}
-
+	/* Other errors cannot be raised, but are printed to stderr */
 	func = PyObject_GetAttrString(mod, "DllCanUnloadNow");
 	Py_DECREF(mod);
 	if (!func) {
-		PyErr_Clear();
+		PyErr_WriteUnraisable(context ? context : Py_None);
 		return E_FAIL;
 	}
 
 	result = PyObject_CallFunction(func, NULL);
 	Py_DECREF(func);
 	if (!result) {
-		PyErr_Clear();
+		PyErr_WriteUnraisable(context ? context : Py_None);
 		return E_FAIL;
 	}
 
 	retval = PyInt_AsLong(result);
-	if (PyErr_Occurred())
-		PyErr_Clear();
+	if (PyErr_Occurred()) {
+		PyErr_WriteUnraisable(context ? context : Py_None);
+		retval = E_FAIL;
+	}
 	Py_DECREF(result);
 	return retval;
 }
@@ -614,21 +479,9 @@ long Call_CanUnloadNow(void)
 STDAPI DllCanUnloadNow(void)
 {
 	long result;
-#ifdef CTYPES_USE_GILSTATE
 	PyGILState_STATE state = PyGILState_Ensure();
-#else
-	if (!Py_IsInitialized()) {
-		LoadPython();
-		LeavePython();
-	}
-	EnterPython();
-#endif
 	result = Call_CanUnloadNow();
-#ifdef CTYPES_USE_GILSTATE
 	PyGILState_Release(state);
-#else
-	LeavePython();
-#endif
 	return result;
 }
 
