@@ -23,9 +23,9 @@
   'inargs', or a completely fresh tuple, depending on several things (is is a
   COM method, are 'paramflags' available).
 
-  3. If [out] parameters are present in paramflags, _build_callargs also
-  creates and returns another object, which is either a single object or a
-  tuple, whcih will later be used to build the function return value.
+  3. _build_callargs also calculates bitarrays containing indexes into
+  the callargs tuple, specifying how to build the return value(s) of
+  the function.
 
   4. _CallProc is then called with the 'callargs' tuple.  _CallProc first
   allocates two arrays.  The first is an array of 'struct argument' items, the
@@ -60,40 +60,52 @@
 #ifdef MS_WIN32
 #include <windows.h>
 #else
-#include <dlfcn.h>
+#include "ctypes_dlfcn.h"
 #endif
 
 #ifdef MS_WIN32
-#define alloca _alloca
+#include <malloc.h>
 #endif
 
 #include <ffi.h>
 #include "ctypes.h"
 
-#ifdef _DEBUG
-#define DEBUG_EXCEPTIONS /* */
+#if defined(_DEBUG) || defined(__MINGW32__)
+/* Don't use structured exception handling on Windows if this is defined.
+   MingW, AFAIK, doesn't support it.
+*/
+#define DONT_USE_SEH
 #endif
 
 #ifdef MS_WIN32
-static char *FormatError(DWORD code)
+PyObject *ComError;
+
+static TCHAR *FormatError(DWORD code)
 {
-	LPVOID lpMsgBuf;
-	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
-		      NULL,
-		      code,
-		      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
-		      (LPSTR) &lpMsgBuf,
-		      0,
-		      NULL);
-	return (char *)lpMsgBuf;
+	TCHAR *lpMsgBuf;
+	DWORD n;
+	n = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
+			  NULL,
+			  code,
+			  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
+			  (LPTSTR) &lpMsgBuf,
+			  0,
+			  NULL);
+	if (n) {
+		while (isspace(lpMsgBuf[n-1]))
+			--n;
+		lpMsgBuf[n] = '\0'; /* rstrip() */
+	}
+	return lpMsgBuf;
 }
 
+#ifndef DONT_USE_SEH
 void SetException(DWORD code, EXCEPTION_RECORD *pr)
 {
-	char *lpMsgBuf;
+	TCHAR *lpMsgBuf;
 	lpMsgBuf = FormatError(code);
 	if(lpMsgBuf) {
-		PyErr_SetString(PyExc_WindowsError, lpMsgBuf);
+		PyErr_SetFromWindowsErr(code);
 		LocalFree(lpMsgBuf);
 	} else {
 		switch (code) {
@@ -246,6 +258,7 @@ static DWORD HandleException(EXCEPTION_POINTERS *ptrs,
 	*record = *ptrs->ExceptionRecord;
 	return EXCEPTION_EXECUTE_HANDLER;
 }
+#endif
 
 static PyObject *
 check_hresult(PyObject *self, PyObject *args)
@@ -447,6 +460,7 @@ struct argument {
  */
 static int ConvParam(PyObject *obj, int index, struct argument *pa)
 {
+	pa->keep = NULL; /* so we cannot forget it later */
 	if (PyCArg_CheckExact(obj)) {
 		PyCArgObject *carg = (PyCArgObject *)obj;
 		pa->ffi_type = carg->pffi_type;
@@ -465,17 +479,17 @@ static int ConvParam(PyObject *obj, int index, struct argument *pa)
 
 	if (PyInt_Check(obj)) {
 		pa->ffi_type = &ffi_type_sint;
-		pa->value.l = PyInt_AS_LONG(obj);
+		pa->value.i = PyInt_AS_LONG(obj);
 		return 0;
 	}
 
 	if (PyLong_Check(obj)) {
 		pa->ffi_type = &ffi_type_sint;
-		pa->value.l = (long)PyLong_AsUnsignedLong(obj);
-		if (pa->value.l == -1 && PyErr_Occurred()) {
+		pa->value.i = (long)PyLong_AsUnsignedLong(obj);
+		if (pa->value.i == -1 && PyErr_Occurred()) {
 			PyErr_Clear();
-			pa->value.l = PyLong_AsLong(obj);
-			if (pa->value.l == -1 && PyErr_Occurred()) {
+			pa->value.i = PyLong_AsLong(obj);
+			if (pa->value.i == -1 && PyErr_Occurred()) {
 				PyErr_SetString(PyExc_OverflowError,
 						"long int too long to convert");
 				return -1;
@@ -543,25 +557,10 @@ static int ConvParam(PyObject *obj, int index, struct argument *pa)
 		}
 		if (PyInt_Check(arg)) {
 			pa->ffi_type = &ffi_type_sint;
-			pa->value.l = PyInt_AS_LONG(arg);
+			pa->value.i = PyInt_AS_LONG(arg);
 			pa->keep = arg;
 			return 0;
 		}
-#if 0
-/* Does this make sense? Now that even Structure and Union types
-   have an _as_parameter_ property implemented in C, which returns
-   a PyCArgObject?
-*/
-		if (CDataObject_Check(arg)) {
-			CDataObject *mem = (CDataObject *)arg;
-			parm->tag = 'V';
-			parm->value.p = mem->b_ptr;
-			parm->size = mem->b_size;
-			/* This consumes the refcount of arg */
-			parm->obj = arg;
-			return parm;
-		}
-#endif
 		Py_DECREF(arg);
 		PyErr_Format(PyExc_TypeError,
 			     "Don't know how to convert parameter %d", index);
@@ -578,18 +577,18 @@ ffi_type *GetType(PyObject *obj)
 	dict = PyType_stgdict(obj);
 	if (dict == NULL)
 		return &ffi_type_sint;
-#ifdef MS_WIN32
+#if defined(MS_WIN32) && !defined(_WIN32_WCE)
 	/* This little trick works correctly with MSVC.
 	   It returns small structures in registers
 	*/
-	if (dict->ffi_type.type == FFI_TYPE_STRUCT) {
-		if (dict->ffi_type.size <= 4)
+	if (dict->ffi_type_pointer.type == FFI_TYPE_STRUCT) {
+		if (dict->ffi_type_pointer.size <= 4)
 			return &ffi_type_sint32;
-		else if (dict->ffi_type.size <= 8)
+		else if (dict->ffi_type_pointer.size <= 8)
 			return &ffi_type_sint64;
 	}
 #endif
-	return &dict->ffi_type;
+	return &dict->ffi_type_pointer;
 }
 
 
@@ -618,8 +617,10 @@ static int _call_function_pointer(int flags,
 	int cc;
 #ifdef MS_WIN32
 	int delta;
-	DWORD dwExceptionCode;
+#ifndef DONT_USE_SEH
+	DWORD dwExceptionCode = 0;
 	EXCEPTION_RECORD record;
+#endif
 #endif
 	/* XXX check before here */
 	if (restype == NULL) {
@@ -629,10 +630,9 @@ static int _call_function_pointer(int flags,
 	}
 	
 	cc = FFI_DEFAULT_ABI;
-#ifdef MS_WIN32
+#if defined(MS_WIN32) && !defined(_WIN32_WCE)
 	if ((flags & FUNCFLAG_CDECL) == 0)
 		cc = FFI_STDCALL;
-	dwExceptionCode = 0;
 #endif
 	if (FFI_OK != ffi_prep_cif(&cif,
 				   cc,
@@ -647,14 +647,14 @@ static int _call_function_pointer(int flags,
 	if ((flags & FUNCFLAG_PYTHONAPI) == 0)
 		Py_UNBLOCK_THREADS
 #ifdef MS_WIN32
-#ifndef DEBUG_EXCEPTIONS
+#ifndef DONT_USE_SEH
 	__try {
 #endif
 		delta =
 #endif
 			ffi_call(&cif, (void *)pProc, resmem, avalues);
 #ifdef MS_WIN32
-#ifndef DEBUG_EXCEPTIONS
+#ifndef DONT_USE_SEH
 	}
 	__except (HandleException(GetExceptionInformation(),
 				  &dwExceptionCode, &record)) {
@@ -665,10 +665,12 @@ static int _call_function_pointer(int flags,
 	if ((flags & FUNCFLAG_PYTHONAPI) == 0)
 		Py_BLOCK_THREADS
 #ifdef MS_WIN32
+#ifndef DONT_USE_SEH
 	if (dwExceptionCode) {
 		SetException(dwExceptionCode, &record);
 		return -1;
 	}
+#endif
 	if (delta < 0) {
 		if (flags & FUNCFLAG_CDECL)
 			PyErr_Format(PyExc_ValueError,
@@ -695,115 +697,53 @@ static int _call_function_pointer(int flags,
 	return 0;
 }
 
-#define RESULT_PASTE_INTO 1
-#define RESULT_CALL_RESTYPE 2
-
 /*
- * Convert the C value in result into an instance described by restype
+ * Convert the C value in result into a Python object, depending on restype.
+ *
+ * - If restype is NULL, return a Python integer.
+ * - If restype is None, return None.
+ * - If restype is a simple ctypes type (c_int, c_void_p), call the type's getfunc,
+ *   pass the result to checker and return the result.
+ * - If restype is another ctypes type, return an instance of that.
+ * - Otherwise, call restype and return the result.
  */
 static PyObject *GetResult(PyObject *restype, void *result, PyObject *checker)
 {
 	StgDictObject *dict;
+	PyObject *retval, *v;
 
-	if (restype == NULL) {
-		return getentry("i")->getfunc(result, sizeof(int));
-	}
+	if (restype == NULL)
+		return PyInt_FromLong(*(int *)result);
 
 	if (restype == Py_None) {
 		Py_INCREF(Py_None);
 		return Py_None;
 	}
 
-	if (PointerTypeObject_Check(restype)) {
-		CDataObject *pd;
-		/* There is no Python api to set the pointer value, so we
-		   create an empty (NULL) pointer, and modify it afterwards.
-		*/
-		pd = (CDataObject *)PyObject_CallFunctionObjArgs(restype, NULL);
-		if (!pd)
-			return NULL;
-		if (!CDataObject_Check(pd)) {
-			Py_DECREF(pd);
-			PyErr_SetString(PyExc_TypeError,
-					"BUG: restype call did not return a CDataObject");
-			return NULL;
-		}
-		/* Even better would be to use the buffer interface */
-		memcpy(pd->b_ptr, result, pd->b_size);
-		return (PyObject *)pd;
-	}
-
-	if (StructTypeObject_Check(restype)) {
-		CDataObject *pd;
-
-		pd = (CDataObject *)PyObject_CallFunctionObjArgs(restype, NULL);
-		if (!pd)
-			return NULL;
-		if (!CDataObject_Check(pd)) {
-			Py_DECREF(pd);
-			PyErr_SetString(PyExc_TypeError,
-					"BUG: restype call did not return a CDataObject");
-			return NULL;
-		}
-		/* Even better would be to use the buffer interface */
-		memcpy(pd->b_ptr, result, pd->b_size);
-		return (PyObject *)pd;
-	}
-
 	dict = PyType_stgdict(restype);
-	if (dict && dict->getfunc) {
-		/* This hack is needed for big endian machines.
-		   Is there another way?
-		 */
-		PyObject *retval;
-		char c;
-		short s;
-		int i;
-#if (SIZEOF_LONG != SIZEOF_INT)
-		long l;
-#endif
-		switch (dict->size) {
-		case 1:
-			c = (char)*(long *)result;
-			retval = dict->getfunc(&c, dict->size);
-			break;
-		case SIZEOF_SHORT:
-			s = (short)*(long *)result;
-			retval = dict->getfunc(&s, dict->size);
-			break;
-		case SIZEOF_INT:
-			i = (int)*(long *)result;
-			retval = dict->getfunc(&i, dict->size);
-			break;
-#if (SIZEOF_LONG != SIZEOF_INT)
-		case SIZEOF_LONG:
-			l = (long)*(long *)result;
-			retval = dict->getfunc(&l, dict->size);
-			break;
-#endif
-		default:
-			retval = dict->getfunc(result, dict->size);
-			break;
-		}
-		if (retval == NULL)
-			return NULL;
-		if (checker == NULL)
-			return retval;
-		{
-			PyObject *v;
-			v = PyObject_CallFunctionObjArgs(checker, retval, NULL);
-			if (v == NULL)
-				_AddTraceback("GetResult", __FILE__, __LINE__-2);
+	if (dict == NULL)
+		return PyObject_CallFunction(restype, "i", *(int *)result);
+
+	if (dict->getfunc && !IsSimpleSubType(restype)) {
+		retval = dict->getfunc(result, dict->size);
+		/* If restype is py_object (detected by comparing getfunc with
+		   O_get), we have to call Py_DECREF because O_get has already
+		   called Py_INCREF.
+		*/
+		if (dict->getfunc == getentry("O")->getfunc) {
 			Py_DECREF(retval);
-			return v;
 		}
-	}
-	if (PyCallable_Check(restype))
-		return PyObject_CallFunction(restype, "i",
-					     *(int *)result);
-	PyErr_SetString(PyExc_TypeError,
-			"Bug: cannot convert result");
-	return NULL; /* to silence the compiler */
+	} else
+		retval = CData_FromBaseObj(restype, NULL, 0, result);
+
+	if (!checker || !retval)
+		return retval;
+
+	v = PyObject_CallFunctionObjArgs(checker, retval, NULL);
+	if (v == NULL)
+		_AddTraceback("GetResult", __FILE__, __LINE__-2);
+	Py_DECREF(retval);
+	return v;
 }
 
 /*
@@ -827,6 +767,8 @@ void Extend_Error_Info(PyObject *exc_class, char *fmt, ...)
 	if (cls_str) {
 		PyString_ConcatAndDel(&s, cls_str);
 		PyString_ConcatAndDel(&s, PyString_FromString(": "));
+		if (s == NULL)
+			goto error;
 	} else
 		PyErr_Clear();
 	msg_str = PyObject_Str(v);
@@ -835,18 +777,86 @@ void Extend_Error_Info(PyObject *exc_class, char *fmt, ...)
 	else {
 		PyErr_Clear();
 		PyString_ConcatAndDel(&s, PyString_FromString("???"));
+		if (s == NULL)
+			goto error;
 	}
 	PyErr_SetObject(exc_class, s);
+error:
 	Py_XDECREF(tp);
 	Py_XDECREF(v);
 	Py_XDECREF(tb);
-	Py_DECREF(s);
+	Py_XDECREF(s);
 }
 
 
 #ifdef MS_WIN32
-#define alloca _alloca
+
+static PyObject *
+GetComError(HRESULT errcode, GUID *riid, IUnknown *pIunk)
+{
+	HRESULT hr;
+	ISupportErrorInfo *psei = NULL;
+	IErrorInfo *pei = NULL;
+	BSTR descr=NULL, helpfile=NULL, source=NULL;
+	GUID guid;
+	DWORD helpcontext=0;
+	LPOLESTR progid;
+	PyObject *obj;
+	TCHAR *text;
+
+	hr = pIunk->lpVtbl->QueryInterface(pIunk, &IID_ISupportErrorInfo, (void **)&psei);
+	if (FAILED(hr))
+		goto failed;
+	hr = psei->lpVtbl->InterfaceSupportsErrorInfo(psei, riid);
+	psei->lpVtbl->Release(psei);
+
+	if (FAILED(hr))
+		goto failed;
+	hr = GetErrorInfo(0, &pei);
+	if (hr != S_OK)
+		goto failed;
+
+	pei->lpVtbl->GetDescription(pei, &descr);
+	pei->lpVtbl->GetGUID(pei, &guid);
+	pei->lpVtbl->GetHelpContext(pei, &helpcontext);
+	pei->lpVtbl->GetHelpFile(pei, &helpfile);
+	pei->lpVtbl->GetSource(pei, &source);
+
+  failed:
+	if (pei)
+		pei->lpVtbl->Release(pei);
+
+	progid = NULL;
+	ProgIDFromCLSID(&guid, &progid);
+
+/* XXX Is COMError derived from WindowsError or not? */
+	text = FormatError(errcode);
+#ifdef _UNICODE
+	obj = Py_BuildValue("iu(uuuiu)",
+#else
+	obj = Py_BuildValue("is(uuuiu)",
 #endif
+			    errcode,
+			    text,
+			    descr, source, helpfile, helpcontext,
+			    progid);
+	if (obj) {
+		PyErr_SetObject(ComError, obj);
+		Py_DECREF(obj);
+	}
+	LocalFree(text);
+
+	if (descr)
+		SysFreeString(descr);
+	if (helpfile)
+		SysFreeString(helpfile);
+	if (source)
+		SysFreeString(source);
+
+	return NULL;
+}
+#endif
+
 /*
  * Requirements, must be ensured by the caller:
  * - argtuple is tuple of arguments
@@ -856,11 +866,14 @@ void Extend_Error_Info(PyObject *exc_class, char *fmt, ...)
  */
 PyObject *_CallProc(PPROC pProc,
 		    PyObject *argtuple,
-		    void *pIunk,
+#ifdef MS_WIN32
+		    IUnknown *pIunk,
+		    GUID *iid,
+#endif
 		    int flags,
-		    PyObject *argtypes, /* misleading name: This is a method,
-					   not a type (the .from_param class
-					   nethod) */
+		    PyObject *argtypes, /* misleading name: This is a tuple of
+					   methods, not types: the .from_param
+					   class methods of the types */
 		    PyObject *restype,
 		    PyObject *checker)
 {
@@ -873,20 +886,23 @@ PyObject *_CallProc(PPROC pProc,
 	PyObject *retval = NULL;
 
 	n = argcount = PyTuple_GET_SIZE(argtuple);
+#ifdef MS_WIN32
 	/* an optional COM object this pointer */
 	if (pIunk)
 		++argcount;
+#endif
 
 	args = (struct argument *)alloca(sizeof(struct argument) * argcount);
 	memset(args, 0, sizeof(struct argument) * argcount);
 	argtype_count = argtypes ? PyTuple_GET_SIZE(argtypes) : 0;
+#ifdef MS_WIN32
 	if (pIunk) {
 		args[0].ffi_type = &ffi_type_pointer;
 		args[0].value.p = pIunk;
 		pa = &args[1];
-	} else {
+	} else
+#endif
 		pa = &args[0];
-	}
 
 	/* Convert the arguments */
 	for (i = 0; i < n; ++i, ++pa) {
@@ -942,8 +958,23 @@ PyObject *_CallProc(PPROC pProc,
 					 rtype, resbuf, argcount))
 		goto cleanup;
 
+#ifdef WORDS_BIGENDIAN
+	/* libffi returns the result in a buffer with sizeof(ffi_arg). This
+	   causes problems on big endian machines, since the result buffer
+	   address cannot simply be used as result pointer, instead we must
+	   adjust the pointer value:
+	 */
+	if (rtype->size < sizeof(ffi_arg))
+		resbuf = (char *)resbuf + sizeof(ffi_arg) - rtype->size;
+#endif
+
 #ifdef MS_WIN32
-	if (flags & FUNCFLAG_HRESULT) {
+	if (iid && pIunk) {
+		if (*(int *)resbuf & 0x80000000)
+			retval = GetComError(*(HRESULT *)resbuf, iid, pIunk);
+		else
+			retval = PyInt_FromLong(*(int *)resbuf);
+	} else if (flags & FUNCFLAG_HRESULT) {
 		if (*(int *)resbuf & 0x80000000)
 			retval = PyErr_SetFromWindowsErr(*(int *)resbuf);
 		else
@@ -958,6 +989,16 @@ PyObject *_CallProc(PPROC pProc,
 }
 
 #ifdef MS_WIN32
+
+#ifdef _UNICODE
+#  define PYBUILD_TSTR "u"
+#else
+#  define PYBUILD_TSTR "s"
+#  ifndef _T
+#    define _T(text) text
+#  endif
+#endif
+
 static char format_error_doc[] =
 "FormatError([integer]) -> string\n\
 \n\
@@ -966,7 +1007,7 @@ given, the return value of a call to GetLastError() is used.\n";
 static PyObject *format_error(PyObject *self, PyObject *args)
 {
 	PyObject *result;
-	char *lpMsgBuf;
+	TCHAR *lpMsgBuf;
 	DWORD code = 0;
 	if (!PyArg_ParseTuple(args, "|i:FormatError", &code))
 		return NULL;
@@ -974,7 +1015,7 @@ static PyObject *format_error(PyObject *self, PyObject *args)
 		code = GetLastError();
 	lpMsgBuf = FormatError(code);
 	if (lpMsgBuf) {
-		result = Py_BuildValue("s", lpMsgBuf);
+		result = Py_BuildValue(PYBUILD_TSTR, lpMsgBuf);
 		LocalFree(lpMsgBuf);
 	} else {
 		result = Py_BuildValue("s", "<no description>");
@@ -990,10 +1031,28 @@ The handle may be used to locate exported functions in this\n\
 module.\n";
 static PyObject *load_library(PyObject *self, PyObject *args)
 {
-	char *name;
+	TCHAR *name;
+	PyObject *nameobj;
+	PyObject *ignored;
 	HMODULE hMod;
-	if (!PyArg_ParseTuple(args, "s:LoadLibrary", &name))
+	if (!PyArg_ParseTuple(args, "O|O:LoadLibrary", &nameobj, &ignored))
 		return NULL;
+#ifdef _UNICODE
+	name = alloca((PyString_Size(nameobj) + 1) * sizeof(WCHAR));
+	{
+		int r;
+		char *aname = PyString_AsString(nameobj);
+		if(!aname)
+			return NULL;
+		r = MultiByteToWideChar(CP_ACP, 0, aname, -1, name, PyString_Size(nameobj) + 1);
+		name[r] = 0;
+	}
+#else
+	name = PyString_AsString(nameobj);
+	if(!name)
+		return NULL;
+#endif
+
 	hMod = LoadLibrary(name);
 	if (!hMod)
 		return PyErr_SetFromWindowsErr(GetLastError());
@@ -1060,7 +1119,10 @@ call_commethod(PyObject *self, PyObject *args)
 
 	result =  _CallProc(lpVtbl[index],
 			    arguments,
+#ifdef MS_WIN32
 			    pIunk,
+			    NULL,
+#endif
 			    FUNCFLAG_HRESULT, /* flags */
 			    argtypes, /* self->argtypes */
 			    NULL, /* self->restype */
@@ -1111,15 +1173,16 @@ static PyObject *py_dl_open(PyObject *self, PyObject *args)
 	/* cygwin doesn't define RTLD_LOCAL */
 	int mode = RTLD_NOW;
 #endif
-	if (!PyArg_ParseTuple(args, "z:dlopen", &name))
+	if (!PyArg_ParseTuple(args, "z|i:dlopen", &name, &mode))
 		return NULL;
-	handle = dlopen(name, mode);
+	mode |= RTLD_NOW;
+	handle = ctypes_dlopen(name, mode);
 	if (!handle) {
 		PyErr_SetString(PyExc_OSError,
-				       dlerror());
+				       ctypes_dlerror());
 		return NULL;
 	}
-	return Py_BuildValue("i", handle);
+	return PyLong_FromVoidPtr(handle);
 }
 
 static PyObject *py_dl_close(PyObject *self, PyObject *args)
@@ -1130,7 +1193,7 @@ static PyObject *py_dl_close(PyObject *self, PyObject *args)
 		return NULL;
 	if (dlclose(handle)) {
 		PyErr_SetString(PyExc_OSError,
-				       dlerror());
+				       ctypes_dlerror());
 		return NULL;
 	}
 	Py_INCREF(Py_None);
@@ -1145,10 +1208,10 @@ static PyObject *py_dl_sym(PyObject *self, PyObject *args)
 
 	if (!PyArg_ParseTuple(args, "is:dlsym", &handle, &name))
 		return NULL;
-	ptr = dlsym(handle, name);
+	ptr = ctypes_dlsym(handle, name);
 	if (!ptr) {
 		PyErr_SetString(PyExc_OSError,
-				       dlerror());
+				       ctypes_dlerror());
 		return NULL;
 	}
 	return Py_BuildValue("i", ptr);
@@ -1175,7 +1238,10 @@ call_function(PyObject *self, PyObject *args)
 
 	result =  _CallProc(func,
 			    arguments,
+#ifdef MS_WIN32
 			    NULL,
+			    NULL,
+#endif
 			    0, /* flags */
 			    NULL, /* self->argtypes */
 			    NULL, /* self->restype */
@@ -1203,7 +1269,10 @@ call_cdeclfunction(PyObject *self, PyObject *args)
 
 	result =  _CallProc(func,
 			    arguments,
+#ifdef MS_WIN32
 			    NULL,
+			    NULL,
+#endif
 			    FUNCFLAG_CDECL, /* flags */
 			    NULL, /* self->argtypes */
 			    NULL, /* self->restype */
@@ -1272,8 +1341,9 @@ byref(PyObject *self, PyObject *obj)
 {
 	PyCArgObject *parg;
 	if (!CDataObject_Check(obj)) {
-		PyErr_SetString(PyExc_TypeError,
-				"expected CData instance");
+		PyErr_Format(PyExc_TypeError,
+			     "byref() argument must be a ctypes instance, not '%s'",
+			     obj->ob_type->tp_name);
 		return NULL;
 	}
 
@@ -1297,20 +1367,25 @@ static PyObject *
 addressof(PyObject *self, PyObject *obj)
 {
 	if (CDataObject_Check(obj))
-		return PyInt_FromLong((long)((CDataObject *)obj)->b_ptr);
+		return PyLong_FromVoidPtr(((CDataObject *)obj)->b_ptr);
 	PyErr_SetString(PyExc_TypeError,
 			"invalid type");
 	return NULL;
 }
 
+static int
+converter(PyObject *obj, void **address)
+{
+	*address = PyLong_AsVoidPtr(obj);
+	return *address != NULL;
+}
+
 static PyObject *
 My_PyObj_FromPtr(PyObject *self, PyObject *args)
 {
-	int i;
 	PyObject *ob;
-	if (!PyArg_ParseTuple(args, "i", &i))
+	if (!PyArg_ParseTuple(args, "O&:PyObj_FromPtr", converter, &ob))
 		return NULL;
-	ob = (PyObject *)i;
 	Py_INCREF(ob);
 	return ob;
 }
@@ -1334,7 +1409,7 @@ My_Py_DECREF(PyObject *self, PyObject *arg)
 #ifdef CTYPES_UNICODE
 
 static char set_conversion_mode_doc[] =
-"FormatError(encoding, errors) -> (previous-encoding, previous-errors)\n\
+"set_conversion_mode(encoding, errors) -> (previous-encoding, previous-errors)\n\
 \n\
 Set the encoding and error handling ctypes uses when converting\n\
 between unicode and strings.  Returns the previous values.\n";
@@ -1345,7 +1420,7 @@ set_conversion_mode(PyObject *self, PyObject *args)
 	char *coding, *mode;
 	PyObject *result;
 
-	if (!PyArg_ParseTuple(args, "zs", &coding, &mode))
+	if (!PyArg_ParseTuple(args, "zs:set_conversion_mode", &coding, &mode))
 		return NULL;
 	result = Py_BuildValue("(zz)", conversion_mode_encoding, conversion_mode_errors);
 	if (coding) {
@@ -1362,152 +1437,8 @@ set_conversion_mode(PyObject *self, PyObject *args)
 }
 #endif
 
-static char cast_doc[] =
-"cast(cobject, ctype) -> ctype-instance\n\
-\n\
-Create an instance of ctype, and copy the internal memory buffer\n\
-of cobject to the new instance.  Should be used to cast one type\n\
-of pointer to another type of pointer.\n\
-Doesn't work correctly with ctypes integers.\n";
-
-static PyObject *cast(PyObject *self, PyObject *args)
-{
-	PyObject *obj, *ctype;
-	struct argument a;
-	CDataObject *result;
-
-	if (!PyArg_ParseTuple(args, "OO", &obj, &ctype))
-		return NULL;
-	if (-1 == ConvParam(obj, 1, &a))
-		return NULL;
-	result = (CDataObject *)PyObject_CallFunctionObjArgs(ctype, NULL);
-	if (result == NULL) {
-		Py_XDECREF(a.keep);
-		return NULL;
-	}
-	// result->b_size
-	// a.ffi_type->size
-	memcpy(result->b_ptr, &a.value,
-	       min(result->b_size, a.ffi_type->size));
-	Py_XDECREF(a.keep);
-	return (PyObject *)result;
-}
-
-static char memmove_doc[] =
-"memmove(dst, src, count) -> adress\n\
-\n\
-Copy count bytes from src to dst, return the dst address as integer.\n";
-
-static PyObject *
-c_memmove(PyObject *self, PyObject *args)
-{
-	struct argument a_dst, a_src;
-	int size;
-	void *c_result;
-	PyObject *result = NULL;
-	PyObject *dst, *src;
-
-	memset(&a_dst, 0, sizeof(struct argument));
-	memset(&a_src, 0, sizeof(struct argument));
-	if (!PyArg_ParseTuple(args, "OOi", &dst, &src, &size))
-		return NULL;
-	if (-1 == ConvParam(dst, 1, &a_dst))
-		goto done;
-	if (-1 == ConvParam(src, 2, &a_src))
-		goto done;
-	c_result = memmove(a_dst.value.p, a_src.value.p, size);
-	result = PyLong_FromVoidPtr(c_result);
-  done:
-	Py_XDECREF(a_dst.keep);
-	Py_XDECREF(a_src.keep);
-	return result;
-}
-
-static char memset_doc[] =
-"memset(dst, c, count) -> adress\n\
-\n\
-Set count bytes starting at dst to c, return the dst address as integer.\n";
-static PyObject *
-c_memset(PyObject *self, PyObject *args)
-{
-	PyObject *dst, *result;
-	struct argument a_dst;
-	void *c_result;
-	int c, count;
-
-	if (!PyArg_ParseTuple(args, "Oii", &dst, &c, &count))
-		return NULL;
-	memset(&a_dst, 0, sizeof(struct argument));
-	if (-1 == ConvParam(dst, 1, &a_dst))
-		return NULL;
-	c_result = memset(a_dst.value.p, c, count);
-	result = PyLong_FromVoidPtr(c_result);
-	Py_XDECREF(a_dst.keep);
-	return result;
-}
-
-static char string_at_doc[] =
-"string_at(addr[, size]) -> string\n\
-\n\
-Return the string at addr.\n";
-
-static PyObject *
-string_at(PyObject *self, PyObject *args)
-{
-	PyObject *result = NULL;
-	PyObject *src;
-	struct argument a_arg;
-	int size;
-
-	if (!PyArg_ParseTuple(args, "O|i", &src, &size))
-		return NULL;
-	memset(&a_arg, 0, sizeof(struct argument));
-	if (-1 == ConvParam(src, 1, &a_arg))
-		return NULL;
-	if (PyTuple_GET_SIZE(args) == 1)
-		result = PyString_FromString(a_arg.value.p);
-	else
-		result = PyString_FromStringAndSize(a_arg.value.p, size);
-	Py_XDECREF(a_arg.keep);
-	return result;
-}
-
-#ifdef CTYPES_UNICODE
-static char wstring_at_doc[] =
-"wstring_at(addr[, size]) -> unicode string\n\
-\n\
-Return the wide string at addr.\n";
-
-static PyObject *
-wstring_at(PyObject *self, PyObject *args)
-{
-	PyObject *result = NULL;
-	PyObject *src;
-	struct argument a_arg;
-	int size;
-
-	if (!PyArg_ParseTuple(args, "O|i", &src, &size))
-		return NULL;
-	memset(&a_arg, 0, sizeof(struct argument));
-	if (-1 == ConvParam(src, 1, &a_arg))
-		return NULL;
-	if (PyTuple_GET_SIZE(args) == 1)
-		result = PyUnicode_FromWideChar(a_arg.value.p, wcslen(a_arg.value.p));
-	else
-		result = PyUnicode_FromWideChar(a_arg.value.p, size);
-	Py_XDECREF(a_arg.keep);
-	return result;
-}
-#endif
-
-
 PyMethodDef module_methods[] = {
-	{"string_at", string_at, METH_VARARGS, string_at_doc},
-	{"memmove", c_memmove, METH_VARARGS, memmove_doc},
-	{"memset", c_memset, METH_VARARGS, memset_doc},
-	{"cast", cast, METH_VARARGS, cast_doc},
 #ifdef CTYPES_UNICODE
-	{"wstring_at", wstring_at, METH_VARARGS, wstring_at_doc},
 	{"set_conversion_mode", set_conversion_mode, METH_VARARGS, set_conversion_mode_doc},
 #endif
 #ifdef MS_WIN32
@@ -1518,7 +1449,8 @@ PyMethodDef module_methods[] = {
 	{"call_commethod", call_commethod, METH_VARARGS },
 	{"_check_HRESULT", check_hresult, METH_VARARGS},
 #else
-	{"dlopen", py_dl_open, METH_VARARGS, "dlopen a library"},
+	{"dlopen", py_dl_open, METH_VARARGS,
+	 "dlopen(name, flag={RTLD_GLOBAL|RTLD_LOCAL}) open a shared library"},
 	{"dlclose", py_dl_close, METH_VARARGS, "dlclose a library"},
 	{"dlsym", py_dl_sym, METH_VARARGS, "find symbol in shared library"},
 #endif
