@@ -2,10 +2,10 @@
 #include "structmember.h"
 
 #include <ffi.h>
-#include "ctypes.h"
 #ifdef MS_WIN32
 #include <windows.h>
 #endif
+#include "ctypes.h"
 
 /******************************************************************/
 /*
@@ -34,7 +34,8 @@ CField_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 PyObject *
 CField_FromDesc(PyObject *desc, int index,
 		int *pfield_size, int bitsize, int *pbitofs,
-		int *psize, int *poffset, int *palign, int pack)
+		int *psize, int *poffset, int *palign,
+		int pack, int big_endian)
 {
 	CFieldObject *self;
 	PyObject *proto;
@@ -59,23 +60,20 @@ CField_FromDesc(PyObject *desc, int index,
 		Py_DECREF(self);
 		return NULL;
 	}
+	if (bitsize /* this is a bitfield request */
+	    && *pfield_size /* we have a bitfield open */
 #ifdef MS_WIN32
-	if (bitsize /* this is a bitfield request */
-	    && *pfield_size /* we have a bitfield open */
 	    && dict->size * 8 == *pfield_size /* MSVC */
-	    && (*pbitofs + bitsize) <= *pfield_size) {
-		/* continue bit field */
-		fieldtype = CONT_BITFIELD;
 #else
-	if (bitsize /* this is a bitfield request */
-	    && *pfield_size /* we have a bitfield open */
-	    && dict->size * 8 <= *pfield_size
+	    && dict->size * 8 <= *pfield_size /* GCC */
+#endif
 	    && (*pbitofs + bitsize) <= *pfield_size) {
 		/* continue bit field */
 		fieldtype = CONT_BITFIELD;
+#ifndef MS_WIN32
 	} else if (bitsize /* this is a bitfield request */
 	    && *pfield_size /* we have a bitfield open */
-	    && dict->size * 8 >= *pfield_size /* MSVC */
+	    && dict->size * 8 >= *pfield_size
 	    && (*pbitofs + bitsize) <= dict->size * 8) {
 		/* expand bit field */
 		fieldtype = EXPAND_BITFIELD;
@@ -128,11 +126,10 @@ CField_FromDesc(PyObject *desc, int index,
 
 	switch (fieldtype) {
 	case NEW_BITFIELD:
-#ifdef IS_BIG_ENDIAN
-		self->size = (bitsize << 16) + *pfield_size - *pbitofs - bitsize;
-#else
-		self->size = (bitsize << 16) + *pbitofs;
-#endif
+		if (big_endian)
+			self->size = (bitsize << 16) + *pfield_size - *pbitofs - bitsize;
+		else
+			self->size = (bitsize << 16) + *pbitofs;
 		*pbitofs = bitsize;
 		/* fall through */
 	case NO_BITFIELD:
@@ -162,21 +159,21 @@ CField_FromDesc(PyObject *desc, int index,
 
 		*pfield_size = dict->size * 8;
 
-#ifdef IS_BIG_ENDIAN
-		self->size = (bitsize << 16) + *pfield_size - *pbitofs - bitsize;
-#else
-		self->size = (bitsize << 16) + *pbitofs;
-#endif
+		if (big_endian)
+			self->size = (bitsize << 16) + *pfield_size - *pbitofs - bitsize;
+		else
+			self->size = (bitsize << 16) + *pbitofs;
+
 		self->offset = *poffset - size; /* poffset is already updated for the NEXT field */
 		*pbitofs += bitsize;
 		break;
 
 	case CONT_BITFIELD:
-#ifdef IS_BIG_ENDIAN
-		self->size = (bitsize << 16) + *pfield_size - *pbitofs - bitsize;
-#else
-		self->size = (bitsize << 16) + *pbitofs;
-#endif
+		if (big_endian)
+			self->size = (bitsize << 16) + *pfield_size - *pbitofs - bitsize;
+		else
+			self->size = (bitsize << 16) + *pbitofs;
+
 		self->offset = *poffset - size; /* poffset is already updated for the NEXT field */
 		*pbitofs += bitsize;
 		break;
@@ -248,16 +245,26 @@ CField_repr(CFieldObject *self)
 	PyObject *result;
 	int bits = self->size >> 16;
 	int size = self->size & 0xFFFF;
-	char *name;
+	const char *name;
 
 	name = ((PyTypeObject *)self->proto)->tp_name;
 
 	if (bits)
-		result = PyString_FromFormat("<Field type=%s, ofs=%d, bits=%d>",
-					     name, self->offset, bits);
+		result = PyString_FromFormat(
+#if (PY_VERSION_HEX < 0x02050000)
+			"<Field type=%s, ofs=%d:%d, bits=%d>",
+#else
+			"<Field type=%s, ofs=%zd:%d, bits=%d>",
+#endif
+			name, self->offset, size, bits);
 	else
-		result = PyString_FromFormat("<Field type=%s, ofs=%d, size=%d>",
-					     name, self->offset, size);
+		result = PyString_FromFormat(
+#if (PY_VERSION_HEX < 0x02050000)
+			"<Field type=%s, ofs=%d, size=%d>",
+#else
+			"<Field type=%s, ofs=%zd, size=%d>",
+#endif
+			name, self->offset, size);
 	return result;
 }
 
@@ -407,17 +414,61 @@ get_ulonglong(PyObject *v, unsigned PY_LONG_LONG *p)
 #  define BIT_MASK(size) ((1LL << NUM_BITS(size))-1)
 #endif
 
-#define GET_BITFIELD(v, size) \
-  if (NUM_BITS(size)) { \
-    v <<= (sizeof(v)*8 - LOW_BIT(size) - NUM_BITS(size)); \
-    v >>= (sizeof(v)*8 - NUM_BITS(size)); \
-  }
+/* This macro CHANGES the first parameter IN PLACE. For proper sign handling,
+   we must first shift left, then right.
+*/
+#define GET_BITFIELD(v, size)						\
+	if (NUM_BITS(size)) {						\
+		v <<= (sizeof(v)*8 - LOW_BIT(size) - NUM_BITS(size));	\
+		v >>= (sizeof(v)*8 - NUM_BITS(size));			\
+	}
 
-#define SET(x, v, size) \
-  NUM_BITS(size) ? \
-  ( ( x & ~(BIT_MASK(size) << LOW_BIT(size)) ) | ( (v & BIT_MASK(size)) << LOW_BIT(size) ) ) \
-  : v
+/* This macro RETURNS the first parameter with the bit field CHANGED. */
+#define SET(x, v, size)							\
+	(NUM_BITS(size) ?						\
+	 ( ( x & ~(BIT_MASK(size) << LOW_BIT(size)) ) | ( (v & BIT_MASK(size)) << LOW_BIT(size) ) ) \
+	 : v)
 
+/* byte swapping macros */
+#define SWAP_2(v)				\
+	( ( (v >> 8) & 0x00FF) |		\
+	  ( (v << 8) & 0xFF00) )
+
+#define SWAP_4(v)			\
+	( ( (v & 0x000000FF) << 24 ) |  \
+	  ( (v & 0x0000FF00) <<  8 ) |  \
+	  ( (v & 0x00FF0000) >>  8 ) |  \
+	  ( ((v >> 24) & 0xFF)) )
+
+#ifdef _MSC_VER
+#define SWAP_8(v)				\
+	( ( (v & 0x00000000000000FFL) << 56 ) |  \
+	  ( (v & 0x000000000000FF00L) << 40 ) |  \
+	  ( (v & 0x0000000000FF0000L) << 24 ) |  \
+	  ( (v & 0x00000000FF000000L) <<  8 ) |  \
+	  ( (v & 0x000000FF00000000L) >>  8 ) |  \
+	  ( (v & 0x0000FF0000000000L) >> 24 ) |  \
+	  ( (v & 0x00FF000000000000L) >> 40 ) |  \
+	  ( ((v >> 56) & 0xFF)) )
+#else
+#define SWAP_8(v)				\
+	( ( (v & 0x00000000000000FFLL) << 56 ) |  \
+	  ( (v & 0x000000000000FF00LL) << 40 ) |  \
+	  ( (v & 0x0000000000FF0000LL) << 24 ) |  \
+	  ( (v & 0x00000000FF000000LL) <<  8 ) |  \
+	  ( (v & 0x000000FF00000000LL) >>  8 ) |  \
+	  ( (v & 0x0000FF0000000000LL) >> 24 ) |  \
+	  ( (v & 0x00FF000000000000LL) >> 40 ) |  \
+	  ( ((v >> 56) & 0xFF)) )
+#endif
+
+#define SWAP_INT SWAP_4
+
+#if SIZEOF_LONG == 4
+# define SWAP_LONG SWAP_4
+#elif SIZEOF_LONG == 8
+# define SWAP_LONG SWAP_8
+#endif
 /*****************************************************************
  * The setter methods return an object which must be kept alive, to keep the
  * data valid which has been stored in the memory block.  The ctypes object
@@ -448,7 +499,7 @@ b_set(void *ptr, PyObject *value, unsigned size)
 	long val;
 	if (get_long(value, &val) < 0)
 		return NULL;
-	*(char *)ptr = (char)SET(*(char *)ptr, (char)val, size);
+	*(signed char *)ptr = (signed char)SET(*(signed char *)ptr, (signed char)val, size);
 	_RET(value);
 }
 
@@ -456,7 +507,7 @@ b_set(void *ptr, PyObject *value, unsigned size)
 static PyObject *
 b_get(void *ptr, unsigned size)
 {
-	char val = *(char *)ptr;
+	signed char val = *(signed char *)ptr;
 	GET_BITFIELD(val, size);
 	return PyInt_FromLong(val);
 }
@@ -485,17 +536,46 @@ static PyObject *
 h_set(void *ptr, PyObject *value, unsigned size)
 {
 	long val;
+	short x;
 	if (get_long(value, &val) < 0)
 		return NULL;
-	*(short *)ptr = (short)SET(*(short *)ptr, (short)val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, (short)val, size);
+	memcpy(ptr, &x, sizeof(x));
 	_RET(value);
 }
 
 
+static PyObject *
+h_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	long val;
+	short field;
+	if (get_long(value, &val) < 0)
+		return NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_2(field);
+	field = SET(field, (short)val, size);
+	field = SWAP_2(field);
+	memcpy(ptr, &field, sizeof(field));
+	_RET(value);
+}
+
 static PyObject *
 h_get(void *ptr, unsigned size)
 {
-	short val = *(short *)ptr;
+	short val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyInt_FromLong((long)val);
+}
+
+static PyObject *
+h_get_sw(void *ptr, unsigned size)
+{
+	short val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_2(val);
 	GET_BITFIELD(val, size);
 	return PyInt_FromLong(val);
 }
@@ -504,10 +584,27 @@ static PyObject *
 H_set(void *ptr, PyObject *value, unsigned size)
 {
 	unsigned long val;
+	unsigned short x;
 	if (get_ulong(value, &val) < 0)
 		return NULL;
-	*(unsigned short *)ptr = (unsigned short)SET(*(unsigned short *)ptr,
-						     (unsigned short)val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, (unsigned short)val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+H_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	unsigned long val;
+	unsigned short field;
+	if (get_ulong(value, &val) < 0)
+		return NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_2(field);
+	field = SET(field, (unsigned short)val, size);
+	field = SWAP_2(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
@@ -515,7 +612,18 @@ H_set(void *ptr, PyObject *value, unsigned size)
 static PyObject *
 H_get(void *ptr, unsigned size)
 {
-	unsigned short val = *(short *)ptr;
+	unsigned short val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyInt_FromLong(val);
+}
+
+static PyObject *
+H_get_sw(void *ptr, unsigned size)
+{
+	unsigned short val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_2(val);
 	GET_BITFIELD(val, size);
 	return PyInt_FromLong(val);
 }
@@ -524,9 +632,27 @@ static PyObject *
 i_set(void *ptr, PyObject *value, unsigned size)
 {
 	long val;
+	int x;
 	if (get_long(value, &val) < 0)
 		return NULL;
-	*(int *)ptr = (int)SET(*(int *)ptr, (int)val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, (int)val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+i_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	long val;
+	int field;
+	if (get_long(value, &val) < 0)
+		return NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_INT(field);
+	field = SET(field, (int)val, size);
+	field = SWAP_INT(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
@@ -534,7 +660,18 @@ i_set(void *ptr, PyObject *value, unsigned size)
 static PyObject *
 i_get(void *ptr, unsigned size)
 {
-	int val = *(int *)ptr;
+	int val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyInt_FromLong(val);
+}
+
+static PyObject *
+i_get_sw(void *ptr, unsigned size)
+{
+	int val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_INT(val);
 	GET_BITFIELD(val, size);
 	return PyInt_FromLong(val);
 }
@@ -567,9 +704,26 @@ static PyObject *
 I_set(void *ptr, PyObject *value, unsigned size)
 {
 	unsigned long val;
+	unsigned int x;
 	if (get_ulong(value, &val) < 0)
 		return  NULL;
-	*(unsigned int *)ptr = (unsigned int)SET(*(unsigned int *)ptr, (unsigned int)val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, (unsigned int)val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+I_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	unsigned long val;
+	unsigned int field;
+	if (get_ulong(value, &val) < 0)
+		return  NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = (unsigned int)SET(field, (unsigned int)val, size);
+	field = SWAP_INT(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
@@ -577,7 +731,18 @@ I_set(void *ptr, PyObject *value, unsigned size)
 static PyObject *
 I_get(void *ptr, unsigned size)
 {
-	unsigned int val = *(unsigned int *)ptr;
+	unsigned int val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyLong_FromUnsignedLong(val);
+}
+
+static PyObject *
+I_get_sw(void *ptr, unsigned size)
+{
+	unsigned int val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_INT(val);
 	GET_BITFIELD(val, size);
 	return PyLong_FromUnsignedLong(val);
 }
@@ -586,9 +751,27 @@ static PyObject *
 l_set(void *ptr, PyObject *value, unsigned size)
 {
 	long val;
+	long x;
 	if (get_long(value, &val) < 0)
 		return NULL;
-	*(long *)ptr = (long)SET(*(long *)ptr, val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+l_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	long val;
+	long field;
+	if (get_long(value, &val) < 0)
+		return NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_LONG(field);
+	field = (long)SET(field, val, size);
+	field = SWAP_LONG(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
@@ -596,7 +779,18 @@ l_set(void *ptr, PyObject *value, unsigned size)
 static PyObject *
 l_get(void *ptr, unsigned size)
 {
-	long val = *(long *)ptr;
+	long val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyInt_FromLong(val);
+}
+
+static PyObject *
+l_get_sw(void *ptr, unsigned size)
+{
+	long val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_LONG(val);
 	GET_BITFIELD(val, size);
 	return PyInt_FromLong(val);
 }
@@ -605,9 +799,27 @@ static PyObject *
 L_set(void *ptr, PyObject *value, unsigned size)
 {
 	unsigned long val;
+	unsigned long x;
 	if (get_ulong(value, &val) < 0)
 		return  NULL;
-	*(unsigned long *)ptr = (unsigned long)SET(*(unsigned long *)ptr, val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+L_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	unsigned long val;
+	unsigned long field;
+	if (get_ulong(value, &val) < 0)
+		return  NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_LONG(field);
+	field = (unsigned long)SET(field, val, size);
+	field = SWAP_LONG(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
@@ -615,7 +827,18 @@ L_set(void *ptr, PyObject *value, unsigned size)
 static PyObject *
 L_get(void *ptr, unsigned size)
 {
-	unsigned long val = *(unsigned long *)ptr;
+	unsigned long val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyLong_FromUnsignedLong(val);
+}
+
+static PyObject *
+L_get_sw(void *ptr, unsigned size)
+{
+	unsigned long val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_LONG(val);
 	GET_BITFIELD(val, size);
 	return PyLong_FromUnsignedLong(val);
 }
@@ -625,16 +848,45 @@ static PyObject *
 q_set(void *ptr, PyObject *value, unsigned size)
 {
 	PY_LONG_LONG val;
+	PY_LONG_LONG x;
 	if (get_longlong(value, &val) < 0)
 		return NULL;
-	*(PY_LONG_LONG *)ptr = (PY_LONG_LONG)SET(*(PY_LONG_LONG *)ptr, val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+q_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	PY_LONG_LONG val;
+	PY_LONG_LONG field;
+	if (get_longlong(value, &val) < 0)
+		return NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_8(field);
+	field = (PY_LONG_LONG)SET(field, val, size);
+	field = SWAP_8(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
 static PyObject *
 q_get(void *ptr, unsigned size)
 {
-	PY_LONG_LONG val = *(PY_LONG_LONG *)ptr;
+	PY_LONG_LONG val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyLong_FromLongLong(val);
+}
+
+static PyObject *
+q_get_sw(void *ptr, unsigned size)
+{
+	PY_LONG_LONG val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_8(val);
 	GET_BITFIELD(val, size);
 	return PyLong_FromLongLong(val);
 }
@@ -643,16 +895,45 @@ static PyObject *
 Q_set(void *ptr, PyObject *value, unsigned size)
 {
 	unsigned PY_LONG_LONG val;
+	unsigned PY_LONG_LONG x;
 	if (get_ulonglong(value, &val) < 0)
 		return NULL;
-	*(unsigned PY_LONG_LONG *)ptr = (unsigned PY_LONG_LONG)SET(*(unsigned PY_LONG_LONG *)ptr, val, size);
+	memcpy(&x, ptr, sizeof(x));
+	x = SET(x, val, size);
+	memcpy(ptr, &x, sizeof(x));
+	_RET(value);
+}
+
+static PyObject *
+Q_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	unsigned PY_LONG_LONG val;
+	unsigned PY_LONG_LONG field;
+	if (get_ulonglong(value, &val) < 0)
+		return NULL;
+	memcpy(&field, ptr, sizeof(field));
+	field = SWAP_8(field);
+	field = (unsigned PY_LONG_LONG)SET(field, val, size);
+	field = SWAP_8(field);
+	memcpy(ptr, &field, sizeof(field));
 	_RET(value);
 }
 
 static PyObject *
 Q_get(void *ptr, unsigned size)
 {
-	unsigned PY_LONG_LONG val = *(unsigned PY_LONG_LONG *)ptr;
+	unsigned PY_LONG_LONG val;
+	memcpy(&val, ptr, sizeof(val));
+	GET_BITFIELD(val, size);
+	return PyLong_FromUnsignedLongLong(val);
+}
+
+static PyObject *
+Q_get_sw(void *ptr, unsigned size)
+{
+	unsigned PY_LONG_LONG val;
+	memcpy(&val, ptr, sizeof(val));
+	val = SWAP_8(val);
 	GET_BITFIELD(val, size);
 	return PyLong_FromUnsignedLongLong(val);
 }
@@ -676,14 +957,48 @@ d_set(void *ptr, PyObject *value, unsigned size)
 			     value->ob_type->tp_name);
 		return NULL;
 	}
-	*(double *)ptr = x;
+	memcpy(ptr, &x, sizeof(double));
 	_RET(value);
 }
 
 static PyObject *
 d_get(void *ptr, unsigned size)
 {
-	return PyFloat_FromDouble(*(double *)ptr);
+	double val;
+	memcpy(&val, ptr, sizeof(val));
+	return PyFloat_FromDouble(val);
+}
+
+static PyObject *
+d_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	double x;
+
+	x = PyFloat_AsDouble(value);
+	if (x == -1 && PyErr_Occurred()) {
+		PyErr_Format(PyExc_TypeError,
+			     " float expected instead of %s instance",
+			     value->ob_type->tp_name);
+		return NULL;
+	}
+#ifdef WORDS_BIGENDIAN
+	if (_PyFloat_Pack8(x, (unsigned char *)ptr, 1))
+		return NULL;
+#else
+	if (_PyFloat_Pack8(x, (unsigned char *)ptr, 0))
+		return NULL;
+#endif
+	_RET(value);
+}
+
+static PyObject *
+d_get_sw(void *ptr, unsigned size)
+{
+#ifdef WORDS_BIGENDIAN
+	return PyFloat_FromDouble(_PyFloat_Unpack8(ptr, 1));
+#else
+	return PyFloat_FromDouble(_PyFloat_Unpack8(ptr, 0));
+#endif
 }
 
 static PyObject *
@@ -698,16 +1013,60 @@ f_set(void *ptr, PyObject *value, unsigned size)
 			     value->ob_type->tp_name);
 		return NULL;
 	}
-	*(float *)ptr = x;
+	memcpy(ptr, &x, sizeof(x));
 	_RET(value);
 }
 
 static PyObject *
 f_get(void *ptr, unsigned size)
 {
-	return PyFloat_FromDouble(*(float *)ptr);
+	float val;
+	memcpy(&val, ptr, sizeof(val));
+	return PyFloat_FromDouble(val);
 }
 
+static PyObject *
+f_set_sw(void *ptr, PyObject *value, unsigned size)
+{
+	float x;
+
+	x = (float)PyFloat_AsDouble(value);
+	if (x == -1 && PyErr_Occurred()) {
+		PyErr_Format(PyExc_TypeError,
+			     " float expected instead of %s instance",
+			     value->ob_type->tp_name);
+		return NULL;
+	}
+#ifdef WORDS_BIGENDIAN
+	if (_PyFloat_Pack4(x, (unsigned char *)ptr, 1))
+		return NULL;
+#else
+	if (_PyFloat_Pack4(x, (unsigned char *)ptr, 0))
+		return NULL;
+#endif
+	_RET(value);
+}
+
+static PyObject *
+f_get_sw(void *ptr, unsigned size)
+{
+#ifdef WORDS_BIGENDIAN
+	return PyFloat_FromDouble(_PyFloat_Unpack4(ptr, 1));
+#else
+	return PyFloat_FromDouble(_PyFloat_Unpack4(ptr, 0));
+#endif
+}
+
+/*
+  py_object refcounts:
+
+  1. If we have a py_object instance, O_get must Py_INCREF the returned
+  object, of course.  If O_get is called from a function result, no py_object
+  instance is created - so callproc.c::GetResult has to call Py_DECREF.
+
+  2. The memory block in py_object owns a refcount.  So, py_object must call
+  Py_DECREF on destruction.  Maybe only when b_needsfree is non-zero.
+*/
 static PyObject *
 O_get(void *ptr, unsigned size)
 {
@@ -719,12 +1078,14 @@ O_get(void *ptr, unsigned size)
 					"PyObject is NULL?");
 		return NULL;
 	}
+	Py_INCREF(ob);
 	return ob;
 }
 
 static PyObject *
 O_set(void *ptr, PyObject *value, unsigned size)
 {
+	/* Hm, does the memory block need it's own refcount or not? */
 	*(PyObject **)ptr = value;
 	Py_INCREF(value);
 	return value;
@@ -927,7 +1288,6 @@ z_set(void *ptr, PyObject *value, unsigned size)
 		if (str == NULL)
 			return NULL;
 		*(char **)ptr = PyString_AS_STRING(str);
-		Py_INCREF(str);
 		return str;
 	} else if (PyInt_Check(value) || PyLong_Check(value)) {
 		*(char **)ptr = (char *)PyInt_AsUnsignedLongMask(value);
@@ -943,9 +1303,17 @@ static PyObject *
 z_get(void *ptr, unsigned size)
 {
 	/* XXX What about invalid pointers ??? */
-	if (*(void **)ptr)
+	if (*(void **)ptr) {
+#if defined(MS_WIN32) && !defined(_WIN32_WCE)
+		if (IsBadStringPtrA(*(char **)ptr, -1)) {
+			PyErr_Format(PyExc_ValueError,
+				     "invalid string pointer %p",
+				     ptr);
+			return NULL;
+		}
+#endif
 		return PyString_FromString(*(char **)ptr);
-	else {
+	} else {
 		Py_INCREF(Py_None);
 		return Py_None;
 	}
@@ -1007,6 +1375,7 @@ Z_set(void *ptr, PyObject *value, unsigned size)
 		if (-1 == PyUnicode_AsWideChar((PyUnicodeObject *)value,
 					       buffer, PyUnicode_GET_SIZE(value))) {
 			Py_DECREF(value);
+			Py_DECREF(keep);
 			return NULL;
 		}
 		Py_DECREF(value);
@@ -1127,26 +1496,26 @@ static struct fielddesc formattable[] = {
 	{ 'b', b_set, b_get, &ffi_type_schar},
 	{ 'B', B_set, B_get, &ffi_type_uchar},
 	{ 'c', c_set, c_get, &ffi_type_schar},
-	{ 'd', d_set, d_get, &ffi_type_double},
-	{ 'f', f_set, f_get, &ffi_type_float},
-	{ 'h', h_set, h_get, &ffi_type_sshort},
-	{ 'H', H_set, H_get, &ffi_type_ushort},
-	{ 'i', i_set, i_get, &ffi_type_sint},
-	{ 'I', I_set, I_get, &ffi_type_uint},
+	{ 'd', d_set, d_get, &ffi_type_double, d_set_sw, d_get_sw},
+	{ 'f', f_set, f_get, &ffi_type_float, f_set_sw, f_get_sw},
+	{ 'h', h_set, h_get, &ffi_type_sshort, h_set_sw, h_get_sw},
+	{ 'H', H_set, H_get, &ffi_type_ushort, H_set_sw, H_get_sw},
+	{ 'i', i_set, i_get, &ffi_type_sint, i_set_sw, i_get_sw},
+	{ 'I', I_set, I_get, &ffi_type_uint, I_set_sw, I_get_sw},
 /* XXX Hm, sizeof(int) == sizeof(long) doesn't hold on every platform */
 /* As soon as we can get rid of the type codes, this is no longer a problem */
 #if SIZEOF_LONG == 4
-	{ 'l', l_set, l_get, &ffi_type_sint},
-	{ 'L', L_set, L_get, &ffi_type_uint},
+	{ 'l', l_set, l_get, &ffi_type_sint, l_set_sw, l_get_sw},
+	{ 'L', L_set, L_get, &ffi_type_uint, L_set_sw, L_get_sw},
 #elif SIZEOF_LONG == 8
-	{ 'l', l_set, l_get, &ffi_type_slong},
-	{ 'L', L_set, L_get, &ffi_type_ulong},
+	{ 'l', l_set, l_get, &ffi_type_slong, l_set_sw, l_get_sw},
+	{ 'L', L_set, L_get, &ffi_type_ulong, L_set_sw, L_get_sw},
 #else
 # error
 #endif
 #ifdef HAVE_LONG_LONG
-	{ 'q', q_set, q_get, &ffi_type_slong},
-	{ 'Q', Q_set, Q_get, &ffi_type_ulong},
+	{ 'q', q_set, q_get, &ffi_type_slong, q_set_sw, q_get_sw},
+	{ 'Q', Q_set, Q_get, &ffi_type_ulong, Q_set_sw, Q_get_sw},
 #endif
 	{ 'P', P_set, P_get, &ffi_type_pointer},
 	{ 'z', z_set, z_get, &ffi_type_pointer},
