@@ -1,5 +1,8 @@
 #include "Python.h"
 #include <ffi.h>
+#ifdef MS_WIN32
+#include <windows.h>
+#endif
 #include "ctypes.h"
 
 /******************************************************************/
@@ -35,8 +38,8 @@ static void
 StgDict_dealloc(StgDictObject *self)
 {
 	StgDict_clear(self);
-	PyMem_Free(self->ffi_type.elements);
-	((PyObject *)self)->ob_type->tp_free((PyObject *)self);
+	PyMem_Free(self->ffi_type_pointer.elements);
+	PyDict_Type.tp_dealloc((PyObject *)self);
 }
 
 int
@@ -46,8 +49,8 @@ StgDict_clone(StgDictObject *dst, StgDictObject *src)
 	int size;
 
 	StgDict_clear(dst);
-	PyMem_Free(dst->ffi_type.elements);
-	dst->ffi_type.elements = NULL;
+	PyMem_Free(dst->ffi_type_pointer.elements);
+	dst->ffi_type_pointer.elements = NULL;
 
 	d = (char *)dst;
 	s = (char *)src;
@@ -61,13 +64,15 @@ StgDict_clone(StgDictObject *dst, StgDictObject *src)
 	Py_XINCREF(dst->restype);
 	Py_XINCREF(dst->checker);
 
-	if (src->ffi_type.elements == NULL)
+	if (src->ffi_type_pointer.elements == NULL)
 		return 0;
 	size = sizeof(ffi_type *) * (src->length + 1);
-	dst->ffi_type.elements = PyMem_Malloc(size);
-	if (dst->ffi_type.elements == NULL)
+	dst->ffi_type_pointer.elements = PyMem_Malloc(size);
+	if (dst->ffi_type_pointer.elements == NULL)
 		return -1;
-	memcpy(dst->ffi_type.elements, src->ffi_type.elements, size);
+	memcpy(dst->ffi_type_pointer.elements,
+	       src->ffi_type_pointer.elements,
+	       size);
 	return 0;
 }
 
@@ -177,6 +182,8 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 	PyObject *isPacked;
 	int pack = 0;
 	int ffi_ofs;
+	int big_endian;
+
 	/* HACK Alert: I cannot be bothered to fix ctypes.com, so there has to
 	   be a way to use the old, broken sematics: _fields_ are not extended
 	   but replaced in subclasses.
@@ -188,6 +195,12 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 	if (fields == NULL)
 		return 0;
 
+#ifdef WORDS_BIGENDIAN
+	big_endian = PyObject_HasAttrString(type, "_swappedbytes_") ? 0 : 1;
+#else
+	big_endian = PyObject_HasAttrString(type, "_swappedbytes_") ? 1 : 0;
+#endif
+
 	use_broken_old_ctypes_semantics = \
 		PyObject_HasAttrString(type, "_use_broken_old_ctypes_structure_semantics_");
 
@@ -206,7 +219,7 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 
 	len = PySequence_Length(fields);
 	if (len == -1) {
-		PyErr_SetString(PyExc_AttributeError,
+		PyErr_SetString(PyExc_TypeError,
 				"'_fields_' must be a sequence of pairs");
 		return -1;
 	}
@@ -219,10 +232,12 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 				"_fields_ is final");
 		return -1;
 	}
+	/* XXX This should probably be moved to a point when all this
+	   stuff is sucessfully finished. */
 	stgdict->flags |= DICTFLAG_FINAL;	/* set final */
 
-	if (stgdict->ffi_type.elements)
-		PyMem_Free(stgdict->ffi_type.elements);
+	if (stgdict->ffi_type_pointer.elements)
+		PyMem_Free(stgdict->ffi_type_pointer.elements);
 
 	basedict = PyType_stgdict((PyObject *)((PyTypeObject *)type)->tp_base);
 	if (basedict && !use_broken_old_ctypes_semantics) {
@@ -230,10 +245,12 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 		align = basedict->align;
 		union_size = 0;
 		total_align = align ? align : 1;
-		stgdict->ffi_type.type = FFI_TYPE_STRUCT;
-		stgdict->ffi_type.elements = PyMem_Malloc(sizeof(ffi_type *) * (basedict->length + len + 1));
-		memset(stgdict->ffi_type.elements, 0, sizeof(ffi_type *) * (basedict->length + len + 1));
-		memcpy(stgdict->ffi_type.elements, basedict->ffi_type.elements,
+		stgdict->ffi_type_pointer.type = FFI_TYPE_STRUCT;
+		stgdict->ffi_type_pointer.elements = PyMem_Malloc(sizeof(ffi_type *) * (basedict->length + len + 1));
+		memset(stgdict->ffi_type_pointer.elements, 0,
+		       sizeof(ffi_type *) * (basedict->length + len + 1));
+		memcpy(stgdict->ffi_type_pointer.elements,
+		       basedict->ffi_type_pointer.elements,
 		       sizeof(ffi_type *) * (basedict->length));
 		ffi_ofs = basedict->length;
 	} else {
@@ -242,9 +259,10 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 		align = 0;
 		union_size = 0;
 		total_align = 1;
-		stgdict->ffi_type.type = FFI_TYPE_STRUCT;
-		stgdict->ffi_type.elements = PyMem_Malloc(sizeof(ffi_type *) * (len + 1));
-		memset(stgdict->ffi_type.elements, 0, sizeof(ffi_type *) * (len + 1));
+		stgdict->ffi_type_pointer.type = FFI_TYPE_STRUCT;
+		stgdict->ffi_type_pointer.elements = PyMem_Malloc(sizeof(ffi_type *) * (len + 1));
+		memset(stgdict->ffi_type_pointer.elements, 0,
+		       sizeof(ffi_type *) * (len + 1));
 		ffi_ofs = 0;
 	}
 
@@ -263,10 +281,17 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 			return -1;
 		}
 		dict = PyType_stgdict(desc);
-		stgdict->ffi_type.elements[ffi_ofs + i] = &dict->ffi_type;
+		if (dict == NULL) {
+			Py_DECREF(pair);
+			PyErr_Format(PyExc_TypeError,
+				     "second item in _fields_ tuple (index %d) must be a C type",
+				     i);
+			return -1;
+		}
+		stgdict->ffi_type_pointer.elements[ffi_ofs + i] = &dict->ffi_type_pointer;
 		dict->flags |= DICTFLAG_FINAL; /* mark field type final */
 		if (PyTuple_Size(pair) == 3) { /* bits specified */
-			switch(dict->ffi_type.type) {
+			switch(dict->ffi_type_pointer.type) {
 			case FFI_TYPE_UINT8:
 			case FFI_TYPE_UINT16:
 			case FFI_TYPE_UINT32:
@@ -302,14 +327,16 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 		if (isStruct) {
 			prop = CField_FromDesc(desc, i,
 					       &field_size, bitsize, &bitofs,
-					       &size, &offset, &align, pack);
+					       &size, &offset, &align,
+					       pack, big_endian);
 		} else /* union */ {
 			size = 0;
 			offset = 0;
 			align = 0;
 			prop = CField_FromDesc(desc, i,
 					       &field_size, bitsize, &bitofs,
-					       &size, &offset, &align, pack);
+					       &size, &offset, &align,
+					       pack, big_endian);
 			union_size = max(size, union_size);
 		}
 		total_align = max(align, total_align);
@@ -335,8 +362,8 @@ StructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
 	/* Adjust the size according to the alignment requirements */
 	size = ((size + total_align - 1) / total_align) * total_align;
 
-	stgdict->ffi_type.alignment = total_align;
-	stgdict->ffi_type.size = size;
+	stgdict->ffi_type_pointer.alignment = total_align;
+	stgdict->ffi_type_pointer.size = size;
 
 	stgdict->size = size;
 	stgdict->align = total_align;
